@@ -79,6 +79,16 @@ extern "C" {
 #define GETDNS_CONTEXT_CODE_TLS_BACKOFF_TIME_TEXT "Change related to getdns_context_set_tls_backoff_time"
 #define GETDNS_CONTEXT_CODE_TLS_CONNECTION_RETRIES 624
 #define GETDNS_CONTEXT_CODE_TLS_CONNECTION_RETRIES_TEXT "Change related to getdns_context_set_tls_connection_retries"
+
+#define GETDNS_CONTEXT_CODE_TRUST_ANCHORS_URL 625
+#define GETDNS_CONTEXT_CODE_TRUST_ANCHORS_URL_TEXT "Change related to getdns_context_set_trust_anchors_url"
+#define GETDNS_CONTEXT_CODE_TRUST_ANCHORS_VERIFY_CA 626
+#define GETDNS_CONTEXT_CODE_TRUST_ANCHORS_VERIFY_CA_TEXT "Change related to getdns_context_set_trust_anchors_verify_ca"
+#define GETDNS_CONTEXT_CODE_TRUST_ANCHORS_VERIFY_EMAIL 627
+#define GETDNS_CONTEXT_CODE_TRUST_ANCHORS_VERIFY_EMAIL_TEXT "Change related to getdns_context_set_trust_anchors_verify_email"
+#define GETDNS_CONTEXT_CODE_APPDATA_DIR 628
+#define GETDNS_CONTEXT_CODE_APPDATA_DIR_TEXT "Change related to getdns_context_set_appdata_dir"
+
 /** @}
   */
 
@@ -87,8 +97,8 @@ extern "C" {
  * \defgroup versions Version values
  *  @{
  */
-#define GETDNS_VERSION "1.1.3"
-#define GETDNS_NUMERIC_VERSION 0x01010300
+#define GETDNS_VERSION "1.2.1"
+#define GETDNS_NUMERIC_VERSION 0x01020100
 #define GETDNS_API_VERSION "December 2015"
 #define GETDNS_API_NUMERIC_VERSION 0x07df0c00
 /** @}
@@ -520,7 +530,7 @@ typedef enum getdns_loglevel_type {
 #define GETDNS_LOG_CRIT_TEXT    "Critical conditions"
 #define GETDNS_LOG_ERR_TEXT     "Error conditions"
 #define GETDNS_LOG_WARNING_TEXT "Warning conditions"
-#define GETDNS_LOG_NOTICE_TEXT  "normal, but significant, condition"
+#define GETDNS_LOG_NOTICE_TEXT  "Normal, but significant, condition"
 #define GETDNS_LOG_INFO_TEXT    "Informational message"
 #define GETDNS_LOG_DEBUG_TEXT   "Debug-level message"
 
@@ -531,12 +541,145 @@ typedef void (*getdns_logfunc_type) (void *userarg, uint64_t log_systems,
     getdns_loglevel_type, const char *, va_list ap);
 
 /**
- *
+ * Register a function that will be called when there is something to log
+ * equally or more severe than the given level for the given system.
+ * @param[in] context The context from which to get the setting
+ * @param[in] userarg A user defined argument to be passed to the
+ *                    log function.
+ * @param[in] system  A bitwise ORed collection of systems for which the log
+ *                    function should be called.  Currently only logging 
+ *                    information about upstream statistics is available;
+ *                    i.e.: GETDNS_LOG_UPSTREAM_STATS
+ * @param[in] level   A severity level.  The log function will be called
+ *                    only for messages with an equal or more severe level.
+ *                    More severe has a lower value.
+ * @param[in] func    The log function to call with the user argument,
+ *                    the system for which the log message, the severity
+ *                    level, a printf style format string and the arguments
+ *                    for the format string, as parameter.
+ * @see getdns_loglevel_type
+ * @see getdns_logfunc_type
+ * @return GETDNS_RETURN_GOOD when successful
+ * @return GETDNS_RETURN_INVALID_PARAMETER when context was NULL.
  */
 getdns_return_t
 getdns_context_set_logfunc(getdns_context *context, void *userarg,
     uint64_t system, getdns_loglevel_type level, getdns_logfunc_type func);
 
+/**
+ * Define the location for storing library specific data.  The location should
+ * be writable for the current user using the application with which the
+ * library is linked.  Currently this is only used for storing data concerning
+ * zero configuration dnssec.
+ * @param[in] context     The context from which to get the setting
+ * @param[in] appdata_dir A user writable location in which the library can
+ *                        store data.  The last element of the path is tried
+ *                        to be created if it does not exist.  When NULL is
+ *                        given, the default location is used which is
+ *                        ${HOME}/.getdns/ on Unix line systems (Linux, BSD's,
+ *                        MacOS), and %AppData%\getnds\ on Windows.
+ * @return GETDNS_RETURN_GOOD when successful
+ * @return GETDNS_RETURN_INVALID_PARAMETER when context was NULL.
+ */
+getdns_return_t
+getdns_context_set_appdata_dir(
+    getdns_context *context, const char *appdata_dir);
+
+/**
+ * Sets the url for the location of the XML file from which to fetch the
+ * trust anchors with Zero configuration DNSSEC.  The url should be for
+ * http, and the file should have the .xml extension.
+ *
+ * Alongside the XML file, also the S/MIME signature that will be used to 
+ * validate the XML file, will be fetched from the url with the .xml extension
+ * replaced by .p7s.
+ * 
+ * When successfully validated with the verify Certificate Authority and the
+ * verify email address, the context will be equipped with the DNSSEC trust
+ * anchors defined within the XML file as described in RFC7958.
+ * The XML file will also be stored together with the .p7s file in the
+ * appdata directory with the names "root-anchors.xml" and "root-anchors.p7s"
+ * respectively.
+ *
+ * When the trust-anchors from the XML file are used, the root DNSKEY is
+ * tracked and stored in the appdata directory too as "root.key"
+ *
+ * Trust anchors from the XML file will be tried when:
+ *   - There were no other trust anchors provided, either by the default
+ *     trust anchor file "/etc/unbound/getdns-root.key", or set with the
+ *     getdns_context_set_dnssec_trust_anchors() function.
+ *   - or the available trust anchors (from the default location or set by
+ *     the application) caused the root DNSKEY rrset to be BOGUS.
+ *
+ * Trust anchors from the XML file will be read from the root-anchors.xml
+ * file in appdata directory and will only be used when validation with the 
+ * S/MIME signatures in root-anchors.p7s succeeds with the verification
+ * Certificate Authority and the verification email address.
+ *
+ * A (new) version of "root-anchors.xml" and "root-anchors.p7s" will be
+ * fetched when:
+ *   - The appdata directory is writeable by the current used, but the 
+ *     "root-anchors.xml" or "root-anchors.p7s" files were not available.
+ *   - or there is a new root DNSKEY RRset (or signature) and it contains
+ *     key_ids which were not in "root-anchors.xml."
+ *
+ * @see getdns_context_get_trust_anchors_url
+ * @see getdns_context_set_trust_anchors_verify_CA
+ * @see getdns_context_set_trust_anchors_verify_email
+ * @param[in] context     The context to configure
+ * @param[in] url         The url including the XML file from which the
+ *                        trust anchors (and the S/MIME signature) will be
+ *                        fetched.  Default is:
+ *                        http://data.iana.org/root-anchors/root-anchors.xml
+ *                        When NULL is given, the default will be used.
+ * @return GETDNS_RETURN_GOOD when successful
+ * @return GETDNS_RETURN_INVALID_PARAMETER when context was NULL.
+ */
+getdns_return_t
+getdns_context_set_trust_anchors_url(getdns_context *context, const char *url);
+
+/**
+ * Sets the public certificate for the Certificate Authority with which to
+ * validate the XML file with the S/MIME signatures fetch from the url
+ * given with the getdns_context_set_trust_anchors_url() function.
+ * @see getdns_context_get_trust_anchors_verify_CA
+ * @see getdns_context_set_trust_anchors_url
+ * @see getdns_context_set_trust_anchors_verify_email
+ * @param[in] context     The context to configure
+ * @param[in] verify_CA   The certificate of the Certificate Authority with
+ *                        which to validate the XML trust anchors.
+ *                        The default is the ICANN Root CA, which is valid
+ *                        till Dec 18 2029.
+ *                        When NULL is given, the default will be used.
+ *                        When an empty string is given, Zero configuration
+ *                        DNSSEC will be disabled.
+ * @return GETDNS_RETURN_GOOD when successful
+ * @return GETDNS_RETURN_INVALID_PARAMETER when context was NULL.
+ */
+getdns_return_t
+getdns_context_set_trust_anchors_verify_CA(
+    getdns_context *context, const char *verify_CA);
+
+/**
+ * Sets the email address for the Subject of the signer's certificate from the
+ * p7s signature file with which to validate the XML file fetched from the url
+ * given with the getdns_context_set_trust_anchors_url() function.
+ * @see getdns_context_get_trust_anchors_verify_email
+ * @see getdns_context_set_trust_anchors_url
+ * @see getdns_context_set_trust_anchors_verify_CA
+ * @param[in] context      The context to configure
+ * @param[in] verify_email Only signatures from this name are allowed.
+ *                         The default dnssec@iana.org.
+ *                         When NULL is given, the default will be used.
+ *                         When an empty string is given, Zero configuration
+ *                         DNSSEC will be disabled.
+ * @return GETDNS_RETURN_GOOD when successful
+ * @return GETDNS_RETURN_INVALID_PARAMETER when context was NULL.
+ */
+getdns_return_t
+getdns_context_set_trust_anchors_verify_email(
+    getdns_context *context, const char *verify_email);
+
 /**
  * Get the current resolution type setting from this context.
  * @see getdns_context_set_resolution_type
@@ -902,6 +1045,96 @@ getdns_return_t
 getdns_context_get_update_callback(getdns_context *context, void **userarg,
     void (**value) (getdns_context *, getdns_context_code_t, void *));
 
+
+/**
+ * Gets the url for the location of the XML file from which to fetch the
+ * trust anchors with Zero configuration DNSSEC.
+ *
+ * Alongside the XML file, also the S/MIME signature that will be used to 
+ * validate the XML file, will be fetched from the url with the .xml extension
+ * replaced by .p7s.
+ * 
+ * When successfully validated with the verify Certificate Authority and the
+ * verify email address, the context will be equipped with the DNSSEC trust
+ * anchors defined within the XML file as described in RFC7958.
+ * The XML file will also be stored together with the .p7s file in the
+ * appdata directory with the names "root-anchors.xml" and "root-anchors.p7s"
+ * respectively.
+ *
+ * When the trust-anchors from the XML file are used, the root DNSKEY is
+ * tracked and stored in the appdata directory too as "root.key"
+ *
+ * Trust anchors from the XML file will be tried when:
+ *   - There were no other trust anchors provided, either by the default
+ *     trust anchor file "/etc/unbound/getdns-root.key", or set with the
+ *     getdns_context_set_dnssec_trust_anchors() function.
+ *   - or the available trust anchors (from the default location or set by
+ *     the application) caused the root DNSKEY rrset to be BOGUS.
+ *
+ * Trust anchors from the XML file will be read from the root-anchors.xml
+ * file in appdata directory and will only be used when validation with the 
+ * S/MIME signatures in root-anchors.p7s succeeds with the verification
+ * Certificate Authority and the verification email address.
+ *
+ * A (new) version of "root-anchors.xml" and "root-anchors.p7s" will be
+ * fetched when:
+ *   - The appdata directory is writeable by the current used, but the 
+ *     "root-anchors.xml" or "root-anchors.p7s" files were not available.
+ *   - or there is a new root DNSKEY RRset (or signature) and it contains
+ *     key_ids which were not in "root-anchors.xml."
+ *
+ * @see getdns_context_set_trust_anchors_url
+ * @see getdns_context_get_trust_anchors_verify_CA
+ * @see getdns_context_get_trust_anchors_verify_email
+ * @param[in]  context     The context to configure
+ * @param[out] url         The url including the XML file, from which the
+ *                         trust anchors (and the S/MIME signature) will be
+ *                         fetched.  Default is:
+ *                         http://data.iana.org/root-anchors/root-anchors.xml
+ * @return GETDNS_RETURN_GOOD when successful
+ * @return GETDNS_RETURN_INVALID_PARAMETER when context was NULL.
+ */
+getdns_return_t
+getdns_context_get_trust_anchors_url(
+    getdns_context *context, const char **url);
+
+/**
+ * Gets the public certificate for the Certificate Authority with which to
+ * validate the XML file with the S/MIME signatures fetch from the url
+ * given with the getdns_context_set_trust_anchors_url() function.
+ * @see getdns_context_set_trust_anchors_verify_CA
+ * @see getdns_context_get_trust_anchors_url
+ * @see getdns_context_get_trust_anchors_verify_email
+ * @param[in]  context    The context to configure
+ * @param[out] verify_CA  The certificate of the Certificate Authority with
+ *                        which to validate the XML trust anchors.
+ *                        The default is the ICANN Root CA, which is valid
+ *                        till Dec 18 2029.
+ * @return GETDNS_RETURN_GOOD when successful
+ * @return GETDNS_RETURN_INVALID_PARAMETER when context was NULL.
+ */
+getdns_return_t
+getdns_context_get_trust_anchors_verify_CA(
+    getdns_context *context, const char **verify_CA);
+
+/**
+ * Gets the email address for the Subject of the signer's certificate from the
+ * p7s signature file with which to validate the XML file fetched from the url
+ * given with the getdns_context_set_trust_anchors_url() function.
+ * @see getdns_context_set_trust_anchors_verify_email
+ * @see getdns_context_get_trust_anchors_url
+ * @see getdns_context_get_trust_anchors_verify_CA
+ * @param[in]  context      The context to configure
+ * @param[out] verify_email Only signatures from this name are allowed.
+ *                          The default dnssec@iana.org.
+ * @return GETDNS_RETURN_GOOD when successful
+ * @return GETDNS_RETURN_INVALID_PARAMETER when context was NULL.
+ */
+getdns_return_t
+getdns_context_get_trust_anchors_verify_email(
+    getdns_context *context, const char **verify_email);
+
+
 /** @}
  */
 
@@ -914,14 +1147,14 @@ getdns_context_get_update_callback(getdns_context *context, void **userarg,
 /**
  * Get the version number of this implementation.
  * @see versions
- * @return The version number as string.  For example "1.1.3".
+ * @return The version number as string.  For example "1.2.1".
  */
 const char *getdns_get_version(void);
 
 /**
  * Get the version number of this implementation as number.
  * @see versions
- * @return The version number as number.  For example 0x01010300.
+ * @return The version number as number.  For example 0x01020100.
  *         - The most significant byte of this uint32_t is the Major version.
  *         - The second most significant byte is the Minor version.
  *         - The third most significant byte the Patch version.
@@ -1685,6 +1918,7 @@ getdns_str2bindata(const char *str, getdns_bindata **bindata);
 getdns_return_t
 getdns_str2int(const char *str, uint32_t *value);
 
+
 /** @}
  */
 
